from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from django.conf import settings

from .models import Mail, MailTemplate, Attachment, TemplateAttachment
from .tasks import send_asynchronous_mail
from .utils import create_and_send_mail
from django.core.exceptions import ImproperlyConfigured


class TemplateAttachmentInline(admin.TabularInline):
    model = TemplateAttachment


@admin.register(Attachment)
class AttachmentAdmin(admin.ModelAdmin):
    list_display = ('id', 'name')
    ordering = ('-time_created',)
    search_fields = ('name', )
    readonly_fields = ('time_created', )


@admin.register(MailTemplate)
class MailTemplateAdmin(admin.ModelAdmin):

    def test_mail_template(self, request, queryset):

        mails_sent = 0

        if not settings.DJANGO_SAAS_TEST_EMAIL_ADDRESS:
            raise ImproperlyConfigured(
                "You need to add DJANGO_SAAS_TEST_EMAIL_ADDRESS=youremailaddress@example.com to test emails.")

        for template in queryset:
            create_and_send_mail(
                template_name=template.name, context={}, to_address=settings.DJANGO_SAAS_TEST_EMAIL_ADDRESS)

            mails_sent += 1

        if mails_sent == 1:
            message_bit = _("1 Mail template was")
        else:
            message_bit = _("%s Mail templates were") % mails_sent
        self.message_user(request, "%s tested" % message_bit)

    test_mail_template.short_description = "Send test mail now"

    list_display = ('name', 'subject')
    search_fields = []
    ordering = ('name',)
    actions = [test_mail_template, ]
    inlines = [TemplateAttachmentInline, ]


@admin.register(Mail)
class MailAdmin(admin.ModelAdmin):

    def send_mail_now(self, request, queryset):

        mails_sent = 0

        for mail in queryset:
            send_asynchronous_mail(str(mail.id), settings.USE_SENDGRID)
            mails_sent += 1

        if mails_sent == 1:
            message_bit = _("1 Mail was")
        else:
            message_bit = _("%s Mails were") % mails_sent
        self.message_user(request, "%s sent" % message_bit)

    send_mail_now.short_description = "Send mail now"

    list_display = ('id', 'time_created', 'from_address', 'to_address', 'cc_address', 'template', 'subject', 'context',)
    search_fields = ['from_address', 'to_address', 'cc_address', 'subject', 'context', ]
    ordering = ('-time_created',)
    list_filter = ('time_created', 'template')

    actions = [send_mail_now, ]

    readonly_fields = (
        'time_created', 'time_sent', 'time_delivered', 'used_backend', 'delivery_mail_id', 'delivery_status')





# -*- coding: utf-8 -*-
from django.conf.urls import url

from . import views

app_name = 'django_saas_email'

urlpatterns = [
    url(
        regex="^Mail/~create/$",
        view=views.MailCreateView.as_view(),
        name='Mail_create',
    ),
    url(
        regex="^Mail/(?P<pk>\d+)/~delete/$",
        view=views.MailDeleteView.as_view(),
        name='Mail_delete',
    ),
    url(
        regex="^Mail/(?P<pk>\d+)/$",
        view=views.MailDetailView.as_view(),
        name='Mail_detail',
    ),
    url(
        regex="^Mail/(?P<pk>\d+)/~update/$",
        view=views.MailUpdateView.as_view(),
        name='Mail_update',
    ),
    url(
        regex="^Mail/$",
        view=views.MailListView.as_view(),
        name='Mail_list',
    ),
    url(
        regex="^MailTemplate/~create/$",
        view=views.MailTemplateCreateView.as_view(),
        name='MailTemplate_create',
    ),
    url(
        regex="^MailTemplate/(?P<pk>\d+)/~delete/$",
        view=views.MailTemplateDeleteView.as_view(),
        name='MailTemplate_delete',
    ),
    url(
        regex="^MailTemplate/(?P<pk>\d+)/$",
        view=views.MailTemplateDetailView.as_view(),
        name='MailTemplate_detail',
    ),
    url(
        regex="^MailTemplate/(?P<pk>\d+)/~update/$",
        view=views.MailTemplateUpdateView.as_view(),
        name='MailTemplate_update',
    ),
    url(
        regex="^MailTemplate/$",
        view=views.MailTemplateListView.as_view(),
        name='MailTemplate_list',
    ),
    url(
        regex="^Attachment/~create/$",
        view=views.AttachmentCreateView.as_view(),
        name='Attachment_create',
    ),
    url(
        regex="^Attachment/(?P<pk>\d+)/~delete/$",
        view=views.AttachmentDeleteView.as_view(),
        name='Attachment_delete',
    ),
    url(
        regex="^Attachment/(?P<pk>\d+)/$",
        view=views.AttachmentDetailView.as_view(),
        name='Attachment_detail',
    ),
    url(
        regex="^Attachment/(?P<pk>\d+)/~update/$",
        view=views.AttachmentUpdateView.as_view(),
        name='Attachment_update',
    ),
    url(
        regex="^Attachment/$",
        view=views.AttachmentListView.as_view(),
        name='Attachment_list',
    ),
    url(
        regex="^TemplateAttachment/~create/$",
        view=views.TemplateAttachmentCreateView.as_view(),
        name='TemplateAttachment_create',
    ),
    url(
        regex="^TemplateAttachment/(?P<pk>\d+)/~delete/$",
        view=views.TemplateAttachmentDeleteView.as_view(),
        name='TemplateAttachment_delete',
    ),
    url(
        regex="^TemplateAttachment/(?P<pk>\d+)/$",
        view=views.TemplateAttachmentDetailView.as_view(),
        name='TemplateAttachment_detail',
    ),
    url(
        regex="^TemplateAttachment/(?P<pk>\d+)/~update/$",
        view=views.TemplateAttachmentUpdateView.as_view(),
        name='TemplateAttachment_update',
    ),
    url(
        regex="^TemplateAttachment/$",
        view=views.TemplateAttachmentListView.as_view(),
        name='TemplateAttachment_list',
    ),
]



# Generated by Django 2.1.2 on 2018-12-11 16:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_saas_email', '0003_auto_20181016_1827'),
    ]

    operations = [
        migrations.AddField(
            model_name='mail',
            name='cc_address',
            field=models.EmailField(blank=True, help_text="The 'cc' field of the email", max_length=254, null=True, verbose_name='CC email address'),
        ),
    ]



from celery import shared_task

from .models import Mail


@shared_task
def send_asynchronous_mail(mail_uuid, sendgrid_api=False):
    """Send an asynchronous mail by the given ID."""
    try:
        mail = Mail.objects.get(id=mail_uuid)
    except Mail.DoesNotExist:
        raise AttributeError("There is no mail with that UUID")

    mail.send(sendgrid_api=sendgrid_api)



# -*- coding: utf-8 -*-
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    UpdateView,
    ListView
)

from .models import Mail, MailTemplate, Attachment, TemplateAttachment


class MailCreateView(CreateView):
    model = Mail


class MailDeleteView(DeleteView):
    model = Mail


class MailDetailView(DetailView):
    model = Mail


class MailUpdateView(UpdateView):
    model = Mail


class MailListView(ListView):
    model = Mail


class MailTemplateCreateView(CreateView):
    model = MailTemplate


class MailTemplateDeleteView(DeleteView):
    model = MailTemplate


class MailTemplateDetailView(DetailView):
    model = MailTemplate


class MailTemplateUpdateView(UpdateView):
    model = MailTemplate


class MailTemplateListView(ListView):
    model = MailTemplate


class AttachmentCreateView(CreateView):
    model = Attachment


class AttachmentDeleteView(DeleteView):
    model = Attachment


class AttachmentDetailView(DetailView):
    model = Attachment


class AttachmentUpdateView(UpdateView):
    model = Attachment


class AttachmentListView(ListView):
    model = Attachment


class TemplateAttachmentCreateView(CreateView):
    model = TemplateAttachment


class TemplateAttachmentDeleteView(DeleteView):
    model = TemplateAttachment


class TemplateAttachmentDetailView(DetailView):
    model = TemplateAttachment


class TemplateAttachmentUpdateView(UpdateView):
    model = TemplateAttachment


class TemplateAttachmentListView(ListView):
    model = TemplateAttachment



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-08-05 12:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

try:
    from django.contrib.postgres.fields import JSONField
except:
    from jsonfield import JSONField

import tinymce.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Mail',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False, unique=True, verbose_name='ID')),
                ('from_address', models.EmailField(help_text="The 'from' field of the email", max_length=254, verbose_name='Sender email address')),
                ('to_address', models.EmailField(help_text="The 'to' field of the email", max_length=254, verbose_name='Recipient email address')),
                ('delivery_mail_id', models.IntegerField(blank=True, editable=False, help_text='The ID is saved after correct sending', null=True, verbose_name='Unique mail sender ID')),
                ('delivery_status', models.IntegerField(blank=True, editable=False, help_text='The Mail sender status', null=True, verbose_name='Status of Mail sender')),
                ('subject', models.CharField(blank=True, help_text='Subject line for a mail', max_length=500, null=True, verbose_name='Email Subject line')),
                ('context', JSONField(blank=True, help_text='JSON dump of context dictionary used to fill in templates', null=True, verbose_name='Data of email context')),
                ('time_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, help_text='When was the mail created?', verbose_name='Creation time')),
                ('time_sent', models.DateTimeField(blank=True, editable=False, help_text='When was the mail send via the backend?', null=True, verbose_name='Sent time')),
                ('time_delivered', models.DateTimeField(blank=True, editable=False, help_text='Actual delivery time by the email backend', null=True, verbose_name='Delivery time')),
                ('used_backend', models.CharField(blank=True, editable=False, help_text='Which email backend was used for sending?', max_length=128, null=True, verbose_name='E-Mail Backend')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='MailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Template name; a short all-lowercase string', max_length=100, unique=True, verbose_name='Template name')),
                ('subject', models.CharField(help_text='A format string like "Hello {}"; required', max_length=200, verbose_name='Email subject line template')),
                ('html_template', tinymce.models.HTMLField(help_text="The HTML template, written with Django's template syntax; required", verbose_name='HTML template (required)')),
                ('text_template', models.TextField(blank=True, default='', help_text='This is an optional field for adding a custom text-only version of this template. If left blank, the plaintext email will be generated dynamically from the HTML when needed.', null=True, verbose_name='Text template (optional)')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='mail',
            name='template',
            field=models.ForeignKey(help_text='The used template', on_delete=django.db.models.deletion.CASCADE, to='django_saas_email.MailTemplate', verbose_name='Used Mail template'),
        ),
    ]



# Generated by Django 2.1.2 on 2019-01-21 12:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_saas_email', '0004_mail_cc_address'),
    ]

    operations = [
        migrations.AddField(
            model_name='mail',
            name='bcc_address',
            field=models.EmailField(blank=True, help_text="The 'bcc' field of the email", max_length=254, null=True, verbose_name='BCC email address'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


def forwards_func(apps, schema_editor):
    # We get the model from the versioned app registry;
    # if we directly import it, it'll be the wrong version
    MailTemplate = apps.get_model("django_saas_email", "MailTemplate")
    db_alias = schema_editor.connection.alias

    html_template = """
    <meta itemprop="name" content="Confirm Email"
      style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; margin: 0;"/>
<table width="100%" cellpadding="0" cellspacing="0"
       style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; margin: 0;">
  <tr
    style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; margin: 0;">
    <td class="content-block"
        style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; vertical-align: top; margin: 0; padding: 0 0 20px;"
        valign="top">
      <strong>Hello {{first_name}} {{last_name}}!</strong>
    </td>
  </tr>
  <tr
    style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; margin: 0;">
    <td class="content-block"
        style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; vertical-align: top; margin: 0; padding: 0 0 20px;"
        valign="top">
      Welcome to django-saas-email :) You can change the template in the MailTemplate admin.
    </td>
  </tr>
  <tr
    style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; margin: 0;">
    <td class="content-block" itemprop="handler" itemscope itemtype="http://schema.org/HttpActionHandler"
        style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; vertical-align: top; margin: 0; padding: 0 0 20px;"
        valign="top">
      <a href="http://www.google.com" class="btn-primary" itemprop="url"
         style="font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif; box-sizing: border-box; font-size: 14px; color: #FFF; text-decoration: none; line-height: 2em; font-weight: bold; text-align: center; cursor: pointer; display: inline-block; border-radius: 5px; text-transform: capitalize; background-color: #348eda; margin: 0; border-color: #348eda; border-style: solid; border-width: 10px 20px;">Go
        to google.com</a>
    </td>
  </tr>
</table>
    """

    MailTemplate.objects.using(db_alias).bulk_create([
        MailTemplate(name="hello", html_template=html_template, subject="Hello {{first_name}} {{last_name}}!"),
    ])


def reverse_func(apps, schema_editor):
    # forwards_func() creates two Country instances,
    # so reverse_func() should delete them.
    MailTemplate = apps.get_model("django_saas_email", "MailTemplate")
    db_alias = schema_editor.connection.alias
    MailTemplate.objects.using(db_alias).filter(name="hello").delete()


class Migration(migrations.Migration):

    dependencies = [
        ('django_saas_email', '0001_initial'),
    ]
    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



from .models import Mail
from .tasks import send_asynchronous_mail


def create_and_send_mail(**kwargs):
    """Helper method to create and send a mail.

    create_and_send_mail(template_name="hello", context={'name': 'Jens'}, to_address="me@jensneuhaus.de")
    """
    mail = Mail.objects.create_mail(**kwargs)
    send_asynchronous_mail.delay(mail.id)



# Generated by Django 2.1.2 on 2018-10-16 23:27

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('django_saas_email', '0002_add_hello_template'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('attached_file', models.FileField(upload_to='email-attachments')),
                ('time_created', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='Creation time')),
            ],
            options={
                'verbose_name': 'attachment',
                'verbose_name_plural': 'attachments',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TemplateAttachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attachment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='templates', to='django_saas_email.Attachment')),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='django_saas_email.MailTemplate')),
            ],
            options={
                'verbose_name': 'Preselected attachment',
                'verbose_name_plural': 'Preselected attachments',
            },
        ),
        migrations.AddField(
            model_name='mail',
            name='text',
            field=models.TextField(blank=True, null=True, verbose_name='Message text (plain)'),
        ),
        migrations.AlterField(
            model_name='mail',
            name='template',
            field=models.ForeignKey(blank=True, help_text='The used template', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_saas_email.MailTemplate', verbose_name='Used Mail template'),
        ),
        migrations.AddField(
            model_name='mail',
            name='selected_attachments',
            field=models.ManyToManyField(blank=True, to='django_saas_email.Attachment'),
        ),
        migrations.AddField(
            model_name='mailtemplate',
            name='preselected_attachments',
            field=models.ManyToManyField(blank=True, through='django_saas_email.TemplateAttachment', to='django_saas_email.Attachment'),
        ),
    ]



# -*- coding: utf-8 -*-
import base64
import json
import uuid
import os

import html2text
import sendgrid
from django.conf import settings
from tinymce import models as tinymce_models

from .logger import logger

try:
    from django.contrib.postgres.fields import JSONField
except:
    from jsonfield import JSONField

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.mail import EmailMessage, EmailMultiAlternatives
from django.core.validators import validate_email
from django.db import models
from django.template import Context, Template
from django.template import engines
from django.template.utils import InvalidTemplateEngineError
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class AbstractMailTemplate(models.Model):
    html_template_file = "django_saas_email/email_base.html"

    name = models.CharField(
        _("Template name"),
        help_text=_("Template name; a short all-lowercase string"),
        max_length=100,
        unique=True,
    )

    subject = models.CharField(
        _("Email subject line template"),
        help_text=_('A format string like "Hello {}"; required'),
        max_length=200,
    )

    # HTML field for the html
    html_template = tinymce_models.HTMLField(
        _("HTML template (required)"),
        help_text=_(
            "The HTML template, written with Django's template syntax; required"
        ),
    )

    text_template = models.TextField(
        _("Text template (optional)"),
        help_text=_(
            "This is an optional field for adding a custom text-only version of this template. "
            "If left blank, the plaintext email will be generated dynamically from the HTML when needed."
        ),
        default="",
        null=True,
        blank=True,
    )

    preselected_attachments = models.ManyToManyField(
        "Attachment", through="TemplateAttachment", blank=True
    )

    class Meta:
        abstract = True

    def __str__(self):
        return "%s" % self.name

    @property
    def backend(self):
        try:
            return self._backend
        except AttributeError:
            try:
                self._backend = engines['email']
            except InvalidTemplateEngineError:
                self._backend = None
            return self._backend

    @staticmethod
    def get_footer():
        """The used footer in the email."""
        return getattr(settings, "DJANGO_SASS_EMAIL_FOOTER", None)

    def render_subject(self, context):
        """Take a list of values (inputs) and format the subject template, returning the subject."""
        return Template(self.subject, engine=self.backend.engine).render(context)

    def render_with_context(self, context):
        """Return a dictionary containing two strings, the HTML and plaintext output.

        This output is generated by filling in the templates using the provided context.
        If the text_template field is empty, it will dynamically generate the text version from the HTML output.
        """

        # Rendering of EMAIL_CONTENT
        email_content_html = Template(self.html_template, engine=self.backend.engine).render(context)

        # Context for HTML Template, including EMAIL_CONTENT
        html_context = {
            "EMAIL_CONTENT": email_content_html,
            "EMAIL_SUBJECT": self.render_subject(context),
            "EMAIL_FOOTER": self.get_footer(),
        }

        if self.backend:
            template = self.backend.get_template(self.html_template_file)
            html_output = template.render(html_context)
        else:
            html_output = render_to_string(self.html_template_file, html_context)

        if self.text_template:
            text_output = Template(self.text_template, engine=self.backend.engine).render(context)
        else:
            text_output = self.html_to_text(email_content_html)

        return {"html": html_output, "text": text_output}

    def html_to_text(self, html_string):
        """A helper method that converts a string containing HTML into a string with plaintext only.

        This is currently done with html2text, which converts HTML into valid Markdown.
        If no text_template exists, make_output() will use this method to generate the text-only output.
        This method should not be called externally.
        """
        h = html2text.HTML2Text()
        return h.handle(html_string)


class MailTemplate(AbstractMailTemplate):
    pass


class AbstractAttachment(models.Model):
    name = models.CharField(max_length=100)
    attached_file = models.FileField(upload_to="email-attachments")
    time_created = models.DateTimeField(
        verbose_name=_("Creation time"), default=timezone.now, editable=False
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _("attachment")
        verbose_name_plural = _("attachments")
        abstract = True


class Attachment(AbstractAttachment):
    pass


class TemplateAttachment(models.Model):
    attachment = models.ForeignKey(
        Attachment, on_delete=models.CASCADE, related_name="templates"
    )
    template = models.ForeignKey(
        MailTemplate, on_delete=models.CASCADE, related_name="attachments"
    )

    class Meta:
        verbose_name = _("Preselected attachment")
        verbose_name_plural = _("Preselected attachments")


class MailManager(models.Manager):
    def create_mail(
        self,
        template_name=None,
        context=None,
        to_address=None,
        from_address=None,
        subject=None,
        selected_attachments=None,
        text=None,
        cc_address=None,
        bcc_address=None,
    ):
        """Create a Mail object with proper validation.

        e.g.
        mail = Mail.objects.create_mail("hello", {'name': 'Jens'},"me@jensneuhaus.de")
        mail.send()
        """

        if (template_name is None) == (text is None):
            raise ValidationError(
                "Exactly one of `text` or `template_name` must be specified"
            )
        if not to_address:
            raise ValidationError("`to_address` must be specified")
        if template_name:
            if not isinstance(template_name, MailTemplate):
                try:
                    template = MailTemplate.objects.get(name__iexact=template_name)
                except MailTemplate.DoesNotExist:
                    raise ValueError(
                        "{} is not a valid Template name".format(template_name)
                    )
            else:
                template = template_name
            if not context:
                raise ValidationError("`context` must be specified when using template")

            try:
                context_string = json.dumps(context)
                context_json = json.loads(context_string)
            except ValueError:
                raise ValueError("The given context is not valid: {}".format(context))

            if not isinstance(context, dict):
                raise ValueError(
                    "The given context is not a dictionary: {}".format(context)
                )
        else:
            template = None
            context_json = None

        if from_address is None:
            from_address = settings.DEFAULT_FROM_EMAIL

        try:
            validate_email(from_address)
        except ValidationError:
            raise ValueError("The given email is not valid: {}".format(from_address))

        try:
            validate_email(to_address)
        except ValidationError:
            raise ValueError("The given email is not valid: {}".format(to_address))

        mail = self.create(
            template=template,
            context=context_json,
            from_address=from_address,
            to_address=to_address,
            cc_address=cc_address,
            bcc_address=bcc_address,
            subject=subject,
            text=text,
        )

        if selected_attachments is not None:
            for attachment in selected_attachments:
                mail.selected_attachments.add(attachment)

        return mail


class AbstractMail(models.Model):

    id = models.UUIDField(
        _("ID"), primary_key=True, unique=True, default=uuid.uuid4, editable=False
    )

    from_address = models.EmailField(
        _("Sender email address"),
        help_text=_("The 'from' field of the email"),
        null=False,
        blank=False,
    )

    to_address = models.EmailField(
        _("Recipient email address"),
        help_text=_("The 'to' field of the email"),
        null=False,
        blank=False,
    )

    cc_address = models.EmailField(
        _("CC email address"),
        help_text=_("The 'cc' field of the email"),
        null=True,
        blank=True
    )

    bcc_address = models.EmailField(
        _("BCC email address"),
        help_text=_("The 'bcc' field of the email"),
        null=True,
        blank=True
    )

    # delivery_service (Sendgrid etc. - should be a CharField with Options)
    delivery_mail_id = models.IntegerField(
        _("Unique mail sender ID"),
        help_text=_("The ID is saved after correct sending"),
        null=True,
        blank=True,
        editable=False,
    )

    # The following should maybe be a CharField - depending on the anymail output
    delivery_status = models.IntegerField(
        _("Status of Mail sender"),
        help_text=_("The Mail sender status"),
        null=True,
        blank=True,
        editable=False,
    )

    template = models.ForeignKey(
        MailTemplate,
        verbose_name=_("Used Mail template"),
        help_text=_("The used template"),
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )

    text = models.TextField(
        verbose_name=_("Message text (plain)"), null=True, blank=True
    )

    subject = models.CharField(
        _("Email Subject line"),
        help_text=_("Subject line for a mail"),
        max_length=500,
        null=True,
        blank=True,
    )

    context = JSONField(
        _("Data of email context"),
        help_text=_("JSON dump of context dictionary used to fill in templates"),
        null=True,
        blank=True,
    )

    time_created = models.DateTimeField(
        _("Creation time"),
        help_text=_("When was the mail created?"),
        default=timezone.now,
        editable=False,
    )

    time_sent = models.DateTimeField(
        _("Sent time"),
        help_text=_("When was the mail send via the backend?"),
        null=True,
        blank=True,
        editable=False,
    )
    time_delivered = models.DateTimeField(
        _("Delivery time"),
        help_text=_("Actual delivery time by the email backend"),
        null=True,
        blank=True,
        editable=False,
    )

    used_backend = models.CharField(
        _("E-Mail Backend"),
        help_text=_("Which email backend was used for sending?"),
        null=True,
        blank=True,
        max_length=128,
        editable=False,
    )

    selected_attachments = models.ManyToManyField(Attachment, blank=True)

    objects = MailManager()

    class Meta:
        abstract = True

    def __str__(self):
        return "%s to %s" % (self.template, self.to_address)

    @classmethod
    def get_extra_context(cls):
        """Placeholder to add extra context."""
        return {}

    def get_context(self):
        return Context({**self.context, **self.get_extra_context()})

    def save(self, *args, **kwargs):
        if (self.template is None) == (self.text is None):
            raise ValidationError(
                "Exactly one of `template` or `text` must be specified"
            )
        super().save(*args, **kwargs)

    def render_mail(self):
        """Check for existing MailTemplate. Text is needed, HTML is optional.

        Returns a dictionary containing subject, text output, and html output.
        """
        if self.template:
            context = self.get_context()

            try:
                mail_template = MailTemplate.objects.get(name=self.template)
            except MailTemplate.DoesNotExist:
                raise ImproperlyConfigured(
                    "No mail template found with name: {}".format(self.template)
                )

            if self.subject:
                rendered_subject = Template(self.subject).render(context)
            else:
                rendered_subject = mail_template.render_subject(context)

            output_dict = mail_template.render_with_context(context)
            output_dict["subject"] = rendered_subject

            return output_dict
        else:
            return {"subject": self.subject, "text": self.text, "html": None}

    def send(self, sendgrid_api=False):
        """Send the mail using data from the Mail object.

        It can be called directly, but is usually called asynchronously with tasks.send_asynchronous_mail.

        sendgrid_api=True uses the sendgrid API directly (with bypassing django-anymail)
        """
        rendered_output = self.render_mail()

        html_content = rendered_output["html"]
        txt_content = rendered_output["text"]
        rendered_subject = rendered_output["subject"]

        if sendgrid_api:

            if not settings.SENDGRID_API_KEY:
                raise ImproperlyConfigured("No SENDGRID_API_KEY set.")

            sg = sendgrid.SendGridAPIClient(apikey=settings.SENDGRID_API_KEY)
            personalizations = []
            personalizations.append({
                "to": [{"email": self.to_address}]
            })
            if self.cc_address:
                personalizations.append({
                    "cc": [{"email": self.cc_address}]
                })
            if self.bcc_address:
                personalizations.append({
                    "bcc": [{"email": self.bcc_address}]
                })
            data = {
                "personalizations": personalizations,
                "from": {"email": self.from_address},
                "subject": rendered_subject,
                "content": [{"type": "text/plain", "value": txt_content}],
            }
            attachments = []
            for attachment in self.selected_attachments.all():
                content = base64.b64encode(attachment.attached_file.read()).decode(
                    "ascii"
                )
                attachments.append(
                    {
                        "content": content,
                        "filename": os.path.basename(attachment.attached_file.name),
                        "disposition": "attachment",
                    }
                )
            if attachments:
                data["attachments"] = attachments

            response = sg.client.mail.send.post(request_body=data)
            logger.debug(
                "Email with UUID {} was sent with Sendgrid API.".format(self.id)
            )
            logger.debug(
                "Response Status Code: {}, Body: {}, Headers: {}".format(
                    response.status_code, response.body, response.headers
                )
            )

            self.used_backend = "Sendgrid ({})".format(response.status_code)

        else:

            if html_content:
                msg = EmailMultiAlternatives(
                    rendered_subject, txt_content, self.from_address, [self.to_address]
                )
                msg.attach_alternative(html_content, "text/html")

            else:
                msg = EmailMessage(
                    rendered_subject, txt_content, self.from_address, [self.to_address]
                )
            for attachment in self.selected_attachments.all():
                msg.attach(
                    attachment.attached_file.name,
                    content=attachment.attached_file.read(),
                )

            msg.send()

            logger.info("Email with UUID {} was sent.".format(self.id))

            self.used_backend = settings.EMAIL_BACKEND

        self.time_sent = timezone.now()
        self.save()


class Mail(AbstractMail):
    pass
